import traceback as tb
from datetime import datetime as dt
from multiprocessing import Pool

import pandas as pd
import psutil
from pydantic import ValidationError

from ipc.utility.helper import build_validator_schema, get_query_results, \
    sample_metadata_validator_construct_args_for_pool
from ipc.validator.settings import REPOS_FOR_SAMPLE_LEVEL_VALIDATION

# Build and save schemas
SCHEMAS = {}
for repo in REPOS_FOR_SAMPLE_LEVEL_VALIDATION:
    print(f'Getting {repo} samples\' schema...')
    desc_df = get_query_results(f'DESCRIBE {repo}.samples')
    print(f'Building validator schema...')
    SCHEMAS[repo] = build_validator_schema(desc_df)


def check_samples_in_repo_for_errors(repos):
    try:
        num_repos = len(repos)
        all_error_dfs = []
        args_for_pool = sample_metadata_validator_construct_args_for_pool(repos)
        ctr = 1
        print(f'Compiling errors...')
        compiled_error_df = pd.DataFrame(columns=['Field, Error Message', 'Error Type', 'Repo', 'Data Id'])
        print(f'\nWill use {psutil.cpu_count()} cores at my disposal...\n')
        for repo in repos:
            tstart = dt.now()
            print(f'Working on repo: {repo} | Count: {ctr}/{num_repos}')
            process_pool = Pool(processes=psutil.cpu_count(logical=False))
            error_df_results = process_pool.starmap(collect_errors_per_sample, args_for_pool[repo])
            print(f'Time Taken: {dt.now() - tstart}\n')
            all_error_dfs.extend(error_df_results)
            ctr += 1
        compiled_error_df = pd.concat([compiled_error_df, pd.concat(all_error_dfs)], ignore_index=True)
        compiled_error_df['Field'] = compiled_error_df['Field'].apply(lambda row: row[0])
        return compiled_error_df
    except Exception:
        print(f'Error when collecting errors: {tb.format_exc()}')
        return pd.DataFrame()


def collect_errors_per_sample(repo, val):
    try:
        schema = SCHEMAS[repo]
        schema(**val)
    except ValidationError as err:
        error_df = pd.DataFrame(err.errors())
        error_df.rename(columns={'loc': 'Field',
                                 'msg': 'Error Message',
                                 'type': 'Error Type'}, inplace=True)
        error_df['Field'] = error_df['Field'].apply(lambda row: row[0])
        error_df['Repo'] = repo
        error_df['Data ID'] = val['data_id']
        return error_df
    return pd.DataFrame(columns=['Field, Error Message', 'Error Type', 'Repo', 'Data Id'])


if __name__ == '__main__':
    from ipc.validator.settings import REPOS_FOR_SAMPLE_LEVEL_VALIDATION
    repos = REPOS_FOR_SAMPLE_LEVEL_VALIDATION
    args_for_pool = sample_metadata_validator_construct_args_for_pool(repos)
    df = collect_errors_per_sample(*args_for_pool['depmap'][0])
    print(df.head())
